# -*- coding: utf-8 -*-

from flask import (
    Flask,
    request,
)
from tuxput.resources import (
    s3_server,
    verify_bucket,
    verify_token,
    list_buckets,
)
from tuxput.config import TuxConfig
import json

__version__ = "0.0.3"


def create_app():
    app = Flask(
        __name__, instance_relative_config=True, static_folder="static"
    )

    config = TuxConfig()
    print("Setting auth_handler to {0}".format(config.CONF_BUCKET))
    auth_handler = s3_server(
        bucket=config.CONF_BUCKET, region=config.S3_REGION
    )

    @app.context_processor
    def add_site_name():
        return {"SITE_TITLE": config.SITE_TITLE}

    @app.context_processor
    def add_app_version():
        return {"VERSION": __version__}

    @app.route("/upload/<path:path>", methods=["POST"])
    def upload(path=""):
        """Attempt to upload an object."""
        token = request.headers.get("X-tuxput-token")
        requested_bucket = request.headers.get("X-tuxput-bucket")

        if path == "" or path == "/":
            return (
                "Forbidden: you must specify a key to upload the object",
                403,
            )

        if requested_bucket is None:
            requested_bucket = config.DEFAULT_BUCKET

        # check that the bucket exists in our config
        # and get a TuxPutBucket obj
        bucket = verify_bucket(
            auth_handler, requested_bucket, config.CONF_FILE
        )
        if bucket is None:
            return (
                "Requested upload bucket not found '%s'" % requested_bucket,
                404,
            )

        # get user object by token
        user = verify_token(auth_handler, token, config.CONF_FILE)
        if user is None:
            return (
                "Token not found",
                401,
            )

        if not user.path_allowed(bucket.name, path):
            return (
                "Forbidden: token lacks write permission for"
                + "{0} on {1}\n".format(path, bucket.name),
                403,
            )

        # If still here, get ready to upload.
        real_path = bucket.real_path(path)
        print("real_path set to {0}".format(real_path))

        s3_handler = s3_server(bucket=bucket.name, region=config.S3_REGION)

        # If the url is to a specific object, check for overwrite perm.
        if s3_handler.key_exists(real_path):
            if not config.ALLOW_UPLOAD_OVERWRITE:
                return (
                    "Conflict: the key '{0}' already exists".format(path),
                    409,
                )

            # sneaky pete tried to overwrite the passwd file.. reuse previous
            # error
            if (
                real_path == config.CONF_FILE
                and bucket.name == config.CONF_BUCKET
            ):
                return (
                    "Permission denied for {0}".format(path),
                    403,
                )

        return s3_handler.create_signed_url(real_path)

    @app.route("/buckets", defaults={"path": ""})
    def buckets_get(path):
        output = list_buckets(auth_handler, config.CONF_FILE)
        return json.dumps(output), 200

    @app.route("/authorizations", defaults={"path": ""})
    def authorizations_get(path):
        """Return the available authorized buckets and patterns for token"""
        token = request.headers.get("X-tuxput-token")

        # get user object by token
        user = verify_token(auth_handler, token, config.CONF_FILE)
        if user is None:
            return (
                "Token not found",
                401,
            )

        return json.dumps(user.authorizations), 200

    @app.route("/", defaults={"path": ""})
    @app.route("/<path:path>")
    def default_get(path):
        return "Request refused, this site is for upload only\n", 200

    return app



# -*- coding: utf-8 -*-


class UnknownTokenError(Exception):
    """401 error when server cannot find the token"""

    def __init__(self, msg="Invalid token"):
        self.msg = msg
        super().__init__(self.msg)


class NoAuthorizationError(Exception):
    """403 error when server won't allow the upload"""

    def __init__(self, msg="No authorization"):
        self.msg = msg
        super().__init__(self.msg)


class FileConflictError(Exception):
    """409 error when file already on the server"""

    def __init__(self, msg="Conflict: file already on server"):
        self.msg = msg
        super().__init__(self.msg)



# -*- coding: utf-8 -*-

import boto3
import botocore.exceptions
import json
from fnmatch import fnmatch


class TuxPutBucket:
    def __init__(self, json_data):
        self.name = json_data["name"]
        self.root = json_data.get("root")

    def real_path(self, key):
        if self.root not in [None, "", "/"]:
            return self.root + "/" + key

        return key


class TuxUser:
    def __init__(self, json_data):
        self.token = json_data["token"]
        self.user = json_data["user"]
        self.authorizations = {}

        for a in json_data["authorizations"]:
            bucket = a["bucket"]
            authorizations = a.get("restricted_paths")
            if authorizations is None:
                authorizations = ["*"]

            self.authorizations[bucket] = authorizations

    def path_allowed(self, bucket, key):
        for b_pattern in self.authorizations:
            if fnmatch(bucket, b_pattern):
                for auth_pattern in self.authorizations[b_pattern]:
                    if fnmatch(key, auth_pattern):
                        return True
                # if no match found in auth_patterns, keep going
                # and see if there's another b_pattern match.

        # we have exhaused all bucket patterns and their auth patterns
        # and found no matches.  No auth.
        return False


class s3_server:
    def __init__(self, bucket, region):
        self.bucket = bucket
        self.region = region
        self.client = boto3.client("s3", region_name=region)

    def key_exists(self, key):
        """Try to fetch the metadata for an object. If the object does not
        exist, head_object will raise an exception. Returns True if the
        object exists
        """
        try:
            head = self.client.head_object(Bucket=self.bucket, Key=key)
            if head.get("ContentLength") == 0:
                # Disregard empty objects
                return False
        except (
            botocore.exceptions.ParamValidationError,
            botocore.exceptions.ClientError,
        ):
            return False
        return True

    def create_signed_url(self, key):
        return self.client.generate_presigned_post(
            self.bucket, key, ExpiresIn=90
        )

    def select_from_json(self, expression, key):
        response = self.client.select_object_content(
            Bucket=self.bucket,
            Key=key,
            ExpressionType="SQL",
            Expression=expression,
            InputSerialization={
                "CompressionType": "NONE",
                "JSON": {
                    "Type": "Document",
                },
            },
            OutputSerialization={"JSON": {"RecordDelimiter": "\n"}},
        )

        for item in response["Payload"]:
            if "Records" in item:
                return [
                    json.loads(x) for x in item["Records"]["Payload"].split()
                ]


# searches the tuxput.json for a bucket entry with the specified name
def verify_bucket(s3_handler, bucket, key):
    exp = (
        """SELECT * from S3Object[*].buckets[*] b where b.name = '%s'"""
        % bucket
    )
    result = s3_handler.select_from_json(exp, key).pop()
    return TuxPutBucket(result)


def verify_token(s3_handler, token, key):
    exp = (
        """SELECT * from S3Object[*].users[*] u where u.token = '%s'""" % token
    )
    result = s3_handler.select_from_json(exp, key).pop()
    print(result)
    return TuxUser(result)


def list_buckets(s3_handler, key):
    exp = """SELECT b.name from S3Object[*].buckets[*] b"""
    results = s3_handler.select_from_json(exp, key)
    return results



# -*- coding: utf-8 -*-


import click
import os
import requests
import sys
import logging
import json
from time import sleep


from tpcli.exceptions import (
    UnknownTokenError,
    NoAuthorizationError,
    FileConflictError,
)

UPLOAD_QUEUE = []
PRESIGNED = []
RESULTS = {}
RETRIES = 1
WAITTIME = 10


def good_response(code):
    if code in [200, 204]:
        return True
    return False


def scan_for_files(target):
    myfiles = []

    if os.path.isdir(target):
        entries = os.walk(target)
        for e in entries:
            (dirname, subdirs, files) = e
            # all all files in this dir
            myfiles.extend([os.path.join(dirname, x) for x in files])
            # scan through the subdirs recursively
            for d in subdirs:
                myfiles.extend(scan_for_files(os.path.join(dirname, d)))
    elif os.path.isfile(target):
        myfiles.append(target)
    else:
        raise Exception(
            "Unknown filetype included.. {0}... skipping".format(target)
        )

    return myfiles


def list_server_buckets(url, token=None):
    headers = {"X-tuxput-token": token}
    full_url = url + "buckets"
    resp = requests.get(full_url, headers=headers)

    return resp.content


def list_server_authorizations(url, token=None):
    headers = {"X-tuxput-token": token}
    full_url = url + "authorizations"
    resp = requests.get(full_url, headers=headers)

    return resp.content


def get_presigned_url(url, token, upload_file, bucket=None):
    retries = RETRIES
    headers = {"X-tuxput-token": token}
    if bucket is not None:
        headers["X-tuxput-bucket"] = bucket
    full_url = url + "/" + upload_file
    while retries > 0:
        retries -= 1

        logging.info(
            "sending req to {0}/{1} -- attempt {2}/{3}".format(
                url, upload_file, retries, RETRIES
            )
        )

        resp = requests.post(full_url, headers=headers)
        logging.debug("<- response {0}".format(resp.status_code))
        # if successful, we can just duck out
        if good_response(resp.status_code):
            logging.info("presigned url successful")
            return resp

        # can't login, so.. bad token.  Let's bail.
        if resp.status_code == 401:
            logging.debug("raising UnknownTokenError")
            raise UnknownTokenError
        # we don't have permission for this upload.
        elif resp.status_code == 403:
            logging.debug("raising NoAuthorizationError")
            raise NoAuthorizationError
        # file already exists
        elif resp.status_code == 409:
            logging.debug("raising FileConflictError")
            raise FileConflictError

        # Any other problem is most likely an internal server error
        # or a another transient problem... hang out for a second and
        # try it again
        logging.info(
            " waiting for retry {0}/{1} in {2}s".format(
                RETRIES - retries, RETRIES, WAITTIME
            )
        )
        sleep(WAITTIME)

    # If we get here, then we've failed all retries and still haven't been
    # successful.  Let's return the last response and let the caller figure it
    # out
    return resp


def do_s3_upload(presigned, upload_file):
    files = {}
    files["file"] = (upload_file, open(upload_file, "rb"))
    pdata = presigned.json()

    logging.info("Posting to S3..")
    resp = requests.post(pdata["url"], data=pdata["fields"], files=files)
    logging.info(" returned {0}".format(resp.status_code))
    return resp


@click.command()
@click.argument("url", nargs=1)
@click.argument("targets", nargs=-1)
@click.option("--dry-run", "-n", default=False, is_flag=True)
@click.option("--token", "-t", required=True, type=str)
@click.option("--bucket", "-b", required=False, type=str)
@click.option(
    "--list_buckets", "-l", required=False, default=False, is_flag=True
)
@click.option(
    "--list_auths", "-A", required=False, default=False, is_flag=True
)
@click.option("--sign-only", "-s", default=False, is_flag=True)
def tuxput_cli(
    url, targets, dry_run, token, bucket, list_buckets, list_auths, sign_only
):

    if list_buckets:
        print(list_server_buckets(url, token))
        return

    if list_auths:
        print(list_server_authorizations(url, token))
        return

    # collect the files we'll be uploading
    for t in targets:
        UPLOAD_QUEUE.extend(scan_for_files(t))

    for f in UPLOAD_QUEUE:
        logging.info("-- Processing {0}".format(f))

        if dry_run:
            logging.debug("dry-run enabled... skipping")
            continue

        presigned = None
        try:
            presigned = get_presigned_url(url, token, f, bucket)
            if not good_response(presigned.status_code):
                # server errors that the timeout/retry couldn't resolve.  Log
                # the fail and move on.
                RESULTS[
                    f
                ] = "Failed to get presigned URL:  {0}{1}... skipping".format(
                    f, presigned.status_code
                )
                continue
            else:
                PRESIGNED.append({f: presigned.json()})
                RESULTS[f] = "presigned URL created '{0}'".format(
                    presigned.status_code
                )
        except requests.exceptions.ConnectionError as e:
            # server's not there, so don't bother with the rest
            logging.error("unable to contact server: {0}".format(e))
            sys.exit(1)
        except UnknownTokenError:
            # token is invalid, so no point trying to upload the others
            logging.error("invalid token")
            sys.exit(1)
        except FileConflictError:
            RESULTS[f] = "{0} file already exists".format(f)
            continue
        except NoAuthorizationError:
            # Not necessarily fatal.  Log the fail and move on.
            RESULTS[f] = "no authorization to get presigned url"
            continue

        if sign_only:
            logging.info("presigned only... skipping do_s3_upload")
            continue

        if presigned is not None:
            upload = do_s3_upload(presigned, f)
            if not good_response(upload.status_code):
                RESULTS[f] = " failed to upload to AWS: {0}: {1}".format(
                    upload.status_code, upload.text
                )
            else:
                RESULTS[f] = " uploaded to AWS: {0}".format(upload.status_code)

    if sign_only:
        print(json.dumps(PRESIGNED, indent=2))

    print("Results\n--------------")
    for x in RESULTS:
        print("{0} -{1}".format(x, RESULTS[x]))


# TODO possibly in version 2.0?
# --pre/-p pre-run command
# --post/-P post-run command
# --exclude/-x  exclude file or pattern from the targets list


if __name__ == "__main__":
    tuxput_cli(auto_envvar_prefix="TUXPUT_CLI")



# -*- coding: utf-8 -*-

import os
from distutils.util import strtobool


class TuxConfig:
    def __init__(self):
        self.config_strings = {}
        self.config_strings["S3_REGION"] = None
        self.config_strings["CONF_BUCKET"] = None
        self.config_strings["DEFAULT_BUCKET"] = None
        self.config_strings["CONF_FILE"] = "tuxput.json"

        self.config_bools = {}
        self.config_bools["ALLOW_UPLOAD_OVERWRITE"] = False
        for config_var in self.config_bools:
            if config_var in os.environ:
                self.config_bools[config_var] = bool(
                    strtobool(os.environ[config_var])
                )

        for config_var in self.config_strings:
            if config_var in os.environ:
                self.config_strings[config_var] = os.environ[config_var]

        assert (
            self.config_strings["CONF_FILE"] is not None
        ), "Required env var CONF_FILE not set"
        assert (
            self.config_strings["CONF_BUCKET"] is not None
        ), "Required env var CONF_BUCKET not set"
        assert (
            self.config_strings["S3_REGION"] is not None
        ), "Required env var S3_REGION not set"

        if "DEFAULT_BUCKET" not in self.config_strings:
            self.config_strings["DEFAULT_BUCKET"] = self.config_strings[
                "CONF_BUCKET"
            ]

    @property
    def S3_REGION(self):
        return self.config_strings["S3_REGION"]

    @property
    def SITE_TITLE(self):
        return self.config_strings["SITE_TITLE"]

    @property
    def ALLOW_UPLOAD_OVERWRITE(self):
        return self.config_bools["ALLOW_UPLOAD_OVERWRITE"]

    @property
    def CONF_BUCKET(self):
        return self.config_strings["CONF_BUCKET"]

    @property
    def CONF_FILE(self):
        return self.config_strings["CONF_FILE"]

    @property
    def DEFAULT_BUCKET(self):
        return self.config_strings["DEFAULT_BUCKET"]



# -*- coding: utf-8 -*-

from setuptools import find_packages, setup

from os import path
import re

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

with open(path.join("tuxput/__init__.py"), encoding="utf8") as f:
    version = re.search(r'__version__ = "(.*?)"', f.read()).group(1)

setup(
    name="tuxput",
    author="Linaro Limited",
    description="The Serverless File Uploader",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/Linaro/tuxput",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Framework :: Flask",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Testing",
        "Topic :: System :: Operating System Kernels :: Linux",
    ],
    version=version,
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        "boto3",
        "Flask",
        "Click",
    ],
    entry_points="""
        [console_scripts]
        tpcli=tpcli:tuxput_cli
    """,
)



# -*- coding: utf-8 -*-

import os
import arrow


def datetimeformat(date_str):
    dt = arrow.get(date_str)
    return dt.humanize()


def file_name(key):
    return os.path.basename(key)


def folder_name(key):
    return os.path.basename(os.path.normpath(key)) + "/"
